program compiler;
{$APPTYPE CONSOLE}
uses
  sysutils,
  symbols in '..\parser\symbols.pas',
  scanner in '..\parser\scanner.pas',
  errors in '..\parser\errors.pas';

type
  TIdentType = (itConstant, itVariable, itProcedure);
  TIdent = record
             Name : ShortString;
             case Kind : TIdentType of
               itConstant: (val:integer);
               itVariable,itProcedure:(level,adr,size:Integer)
           end;

  TIdentList = Array of TIdent;

  TOpCode = (lit, opr, lod, sto, cal, int, jmp, jpc, wri);
  Instruction = packed record
                  f : TOpCode; //Befehl
                  l : Byte;    //Level
                  a:  Integer; //Adresse
                end;

  TInstructions = Array of Instruction;

var
  Table : TIdentList;
  Code : TInstructions;

  //Code Position
  cx : Integer;

  procedure Expect(Expected : TSymbol);
  begin
    if Symbol <> Expected then
      ErrorExpected([Expected], Symbol)
  end;

  procedure GenCode(f : TOpCode; l, a : Integer);
  begin (*GenCode*)
    if cx > Length(Code) - 1 then
      SetLength(Code, Length(Code) + 64);
    Code[cx].f := f;
    Code[cx].a := a;
    Code[cx].l := l;
    Inc(cx)
  end; (*GenCode*)

  procedure Module;

    function Position(ID : ShortString; TablePosition : Integer) : Integer;
    var
      I : Integer;
    begin
      Table[0].Name := ID;
      I := TablePosition;
      while Table[I].Name <> ID do
        Dec(I);
      Result := I
    end;

    procedure StatementSequence(TablePosition, lev : Integer);
    (*  StatementSeq  = [Statement {";" Statement}]*)
      procedure Statement;
      (*Statement     = ( Ident ":=" (Ident | Expression) |
                  Ident | (WRITE Expression) |
                  ( BEGIN StatementSeq END ) |
                  ( IF Condition THEN StatementSeq END
                    {ELSEIF Condition THEN StatementSeq END }
                    [ELSE StatementSeq END] ) |
                  ( WHILE Condition DO StatementSeq END ) | ";". *)
        procedure Expression;
        (*Expression      = [Vorzeichen] Term {Vorzeichen Term}.*)
          procedure Term;
          (*Term            = Factor {"*"|"/" Factor}.*)
            procedure Factor;
            (*Factor          = Ident | Zahl | ("(" Expression ")").*)
            var
              IdentPos : Integer;
            begin (*Factor*)
              if (Symbol = sIdent) then
              begin
                IdentPos := Position(ID, TablePosition);
                if (IdentPos = 0) then
                  Error(ERR_PARSE_UNKNOWN_IDENT);
                if Table[IdentPos].Kind = itProcedure then
                  Error(ERR_PARSE_VAR_CONSTANT_EXPECTED);

                case Table[IdentPos].Kind of
                  itConstant: GenCode(lit, 0, Table[IdentPos].val);
                  itVariable: GenCode(lod, lev-Table[IdentPos].Level, Table[IdentPos].adr);
                end;

                GetSym
              end
              else if (Symbol = sInteger) then
              begin
                GenCode(lit, 0, num);
                GetSym
              end
              else if (Symbol = sOpenBracket) then
              begin
                GetSym;
                Expression;
                Expect(sCloseBracket);
                GetSym
              end
              else
                ErrorExpected([sIdent, sInteger, sOpenBracket], Symbol)
            end;  (*Factor*)

          var
            Operation : TSymbol;
          begin(*Term*)
            Factor;
            while Symbol in [sStar, sSlash] do
            begin
              Operation := Symbol;
              GetSym;
              Factor;
              case Operation of
                sStar: GenCode(opr, 0, 4);
                sSlash: GenCode(opr, 0, 5);
              end
            end
          end;(*Term*)

        var
          Operation : TSymbol;
        begin (*Expresion*)
          if (Symbol in[sPlus, sMinus]) then
          begin
            Operation := Symbol;
            GetSym;
            Term;
            if Operation = sMinus then
              GenCode(opr, 0, 1)
          end
          else Term;
          while (Symbol in[sPlus, sMinus]) do
          begin
            Operation := Symbol;
            GetSym;
            Term;
            case Operation of
              sPlus: GenCode(opr, 0, 2);
              sMinus: GenCode(opr, 0, 3);
            end
          end
        end; (*Expression*)

        procedure Condition;
        (*Condition       = Expression ("="|"<"|">"|">="|"<="|"#") Expression.*)
        var
          Operation : TSymbol;
        begin (*Condition*)
          Expression;
          Operation := Symbol;
          GetSym;
          Expression;
          case Operation of
            sEqual : GenCode(opr, 0, 8);
            sSmaller : GenCode(opr, 0, 10);
            sBigger : GenCode(opr, 0, 11);
            sBiggerEqual : GenCode(opr, 0, 12);
            sSmallerEqual : GenCode(opr, 0, 13);
            sUnEqual : GEnCode(opr, 0, 9);
          else
            ErrorExpected([sEqual, sSmaller, sBigger, sBiggerEqual, sSmallerEqual, sUnequal], Symbol)
          end
        end; (*Condition*)

      var
        IdentPos : Integer;
        Ident : ShortString;

        CodePosition1, CodePosition2 : Integer;

      begin (*Statement*)
        case Symbol of
          sIdent: begin
                    Ident := Id;
                    IdentPos := Position(ID, TablePosition);
                    if (IdentPos = 0) then
                      Error(ERR_PARSE_UNKNOWN_IDENT);
                    if Table[IdentPos].Kind = itProcedure then
                    begin
                      //Prozeduraufruf
                      GenCode(cal, lev - Table[IdentPos].level, Table[IdentPos].adr);
                      GetSym
                    end else if Table[IdentPos].Kind = itVariable then
                    begin
                      GetSym;
                      Expect(sBecomes);
                      GetSym;
                      Expression;
                      GenCode(sto, lev- Table[IdentPos].level, Table[IdentPos].Adr)
                    end
                    else
                      Error(ERR_PARSE_NO_CONST_ALLOWED)
                  end;
          sWrite: begin
                    GetSym;
                    Expression;
                    GenCode(wri, 0, 0)
                  end;
          sIf :   begin
                    GetSym;
                    Condition;
                    Expect(sThen);
                    GetSym;

                    CodePosition1 := cx;
                    GenCode(jpc, 0, 0);

                    StatementSequence(TablePosition, lev);
                    Expect(sEnd);
                    GetSym;
                    CodePosition2 := cx;
                    GenCode(jmp, 0,0);

                    Code[CodePosition1].a := cx;

                    while Symbol = sElseIf do
                    begin
                      GetSym;
                      Condition;
                      Expect(sThen);
                      GetSym;

                      CodePosition1 := cx;
                      GenCode(jpc, 0, 0);

                      StatementSequence(TablePosition, lev);
                      Expect(sEnd);
                      GetSym;

                      Code[CodePosition2].a := cx;
                      CodePosition2 := cx;
                      GenCode(jmp, 0,0);

                      Code[CodePosition1].a := cx
                    end;
                    if Symbol = sElse then
                    begin
                      GetSym;
                      StatementSequence(TablePosition, lev);
                      Expect(sEnd);
                      GetSym
                    end;
                    Code[CodePosition2].a := cx
                  end;
          sWhile :
                  begin
                    GetSym;
                    CodePosition2 := cx;
                    Condition;
                    Expect(sDo);
                    GetSym;

                    CodePosition1 := cx;
                    GenCode(jpc, 0, 0);

                    StatementSequence(TablePosition, lev);
                    Expect(sEnd);
                    GenCode(jmp, 0, CodePosition2);

                    GetSym;

                    Code[CodePosition1].a := cx
                  end;
          sBegin :
                  begin
                    GetSym;
                    StatementSequence(TablePosition, lev);
                    Expect(sEnd);
                    GetSym
                  end;

        else
          (*Der Fehler darf getrost ignoriert werden*)
          //Error(ERR_PARSE_UNALLOWED_STATEMENT)
        end
      end; (*Statement*)

    begin (*Statement Sequence*)
      Statement;
      while Symbol = sSemiColon do
      begin
        GetSym;
        Statement
      end
    end; (*Statement Sequence*)

    function Declarations(TablePosition : Integer; lev : Integer): Integer;
    var
      DataPos : Integer;
      InitTablePos : Integer;
      InitCodePos : Integer;

    (*
      VarSeq          = VAR VarDecl {";" VarDecl}";".
      ConstSeq        = CONST ConstDecl {";" ConstDecl}";".
      DeclSeq         = {VarSeq | ConstSeq | ProcedureDecl}.

    *)
      procedure Enter(Typ : TIdentType);
      begin
        Inc(TablePosition);
        if TablePosition > Length(Table) - 1 then
          SetLength(Table, Length(Table)+16);
        with Table[TablePosition] do
        begin
          Name := ID;
          Kind := Typ;
          case Kind of
            itConstant: val := num;
            itVariable: begin
                          level := lev;
                          adr := DataPos;
                          Inc(DataPos)
                        end;
            itProcedure: level := lev
          end
        end
      end; (*Enter*)


      procedure ProcedureDecl;
      (*ProcedureDecl   = PROCEDURE Ident";"
                          DeclSeq BEGIN StatementSeq END Ident ";".*)
      var
        ProcedureName : String;
        ProcTablePos : Integer;
      begin
        Expect(sProcedure);
        GetSym;
        Expect(sIdent);
        Enter(itProcedure);
        ProcedureName := ID;

        GetSym;
        Expect(sSemiColon);
        GetSym;
        ProcTablePos := Declarations(TablePosition, lev+1);
        Expect(sBegin);
        GetSym;
        StatementSequence(ProcTablePos, lev +1);
        Expect(sEnd);
        GetSym;
        Expect(sIdent);
        if ProcedureName <> ID then
          Error(Format(ERR_PARSE_WRONG_PROCEDURE_ENDED, [ProcedureName, ID]));
        GetSym;
        Expect(sSemiColon);
        //rcksprung
        GenCode(Opr, 0,0);
        GetSym
      end;

      procedure ConstDecl;
      (*ConstDecl       = Ident "=" Zahl.*)
      begin (*ConstDecl*)
        Expect(sIdent);
        GetSym;
        Expect(sEqual);
        GetSym;
        Expect(sInteger);
        Enter(itConstant);
        GetSym
      end; (*ConstDecl*)

      procedure VarDecl;
      (*VarDecl         = Ident {"," Ident}.*)
      begin (*VarDecl*)
        Expect(sIdent);
        Enter(itVariable);
        GetSym;
        while Symbol = sComma do
        begin
          GetSym;
          Expect(sIdent);
          Enter(itVariable);
          GetSym
        end
      end; (*VarDecl*)

    begin (*Declarations*)
      DataPos := 3;
      InitTablePos := TablePosition;
      InitCodePos := cx;
      Table[TablePosition].adr := cx;

      GenCode(jmp, 0,0);
      while Symbol in [sVar, sConst, sProcedure] do
        case Symbol of
          sVar   : begin
                     GetSym;
                     VarDecl;
                     Expect(sSemiColon);
                     GetSym;
                     while Symbol = sIdent do
                     begin
                       VarDecl;
                       Expect(sSemiColon);
                       GetSym
                     end
                   end;
         sConst  : begin
                     GetSym;
                     ConstDecl;
                     Expect(sSemiColon);
                     GetSym;
                     while Symbol = sIdent do
                     begin
                       ConstDecl;
                       Expect(sSemiColon);
                       GetSym
                     end
                   end;
          sProcedure : ProcedureDecl;
        end;
      Code[table[InitTablePos].adr].a := cx;
      with Table[InitTablePos] do
      begin
        adr := cx;
        size := DataPos
      end;
      //Speicher reservieren
      GenCode(int, 0, DataPos);
      Result := TablePosition
    end; (*Declarations*)

  (*Module        = MODULE Ident ";"
                    DeclSeq
                    BEGIN
                      StatementSeq
                    END Ident ".".*)
  var
    TablePosition : Integer;
    ModuleName : ShortString;

  begin (*Module*)
    Expect(sModule);
    GetSym;
    Expect(sIdent);
    ModuleName := Id;
    GetSym;
    Expect(sSemiColon);
    GetSym;
    TablePosition := Declarations(0,0);
    //Code[0].a := cx;
    Expect(sBegin);
    GetSym;
    StatementSequence(TablePosition, 0);
    Expect(sEnd);

    //Und Ende...
    GenCode(jmp, 0,0);

    GetSym;
    Expect(sIdent);
    if ModuleName <> Id then
      Writeln(Format('%d: Warning: Module ID <> End ID. Code already generated', [Line]));
    GetSym;
    Expect(sSemiColon);
    GetSym;
    if Symbol <> sNone then
      WriteLn(Format('%d: Code after Module END is ignored!', [Line]))
  end (*Module*);

var
  F : File of Instruction;
  I : Integer;
begin
  AssignFile(InFile, ParamStr(1));
  Reset(InFile, 1);
  //Ch mit leerzeichen initialisieren
  Ch := ' ';
  Line := 1;
  cx := 0;
  SetLength(Table, 1);
  GetSym;
  Module;
  AssignFile(F, ParamStr(1)+'.bin');
  ReWrite(F);
  i := 0;
  while i < cx do
  begin
    Write(F, Code[i]);
    Inc(i)
  end;
  CloseFile(F);
  WriteLn('Done, no syntax errors detected... Code successfully generated');
  WriteLn(Format('# Instructions: %d Codesize: %d', [cx, (cx) * SizeOf(Instruction)]));
  ReadLn;
  CloseFile(InFile)
end.
