program interpreter;
{$APPTYPE CONSOLE}
uses
  sysutils;

type
  TOpCode = (lit, opr, lod, sto, cal, int, jmp, jpc, wri);
  Instruction = packed record
                  f : TOpCode; //Befehl
                  l : Byte;    //Parameter 1
                  a : Integer; //Parameter 2
                end;

  TInstructions = Array[0..0] of Instruction;
  PInstructions = ^TInstructions;


var
  Instructions : PInstructions;

  procedure Emulate;
  const
    StackSize = 1024;

  var
    p,b,t : Integer;
    i : Instruction;
    s : Array[1..StackSize] of Integer;

    function Base(I : Integer): Integer;
    var
      b1 : Integer;
    begin
      b1 := b;
      while I > 0 do
      begin
        b1 := s[b1];
        Dec(I)
      end;
      base := b1
    end;

  begin
    WriteLn('Interpreting Code');
    t := 0;
    b := 1;
    p := 0;
    s[1] := 0;
    s[2] := 0;
    s[3] := 0;
    repeat
      //bereichsprfung ist wegen dem Array[0..] nicht angebracht
      {$R-}
      i := Instructions^[p];
      {$R+}
      Inc(p);
      with i do
        case f of
          lit : begin
                  Inc(t);
                  s[t] := a
                end;
          lod : begin
                  Inc(t);
                  s[t] := s[base(l)+a]
                end;
          sto : begin
                  s[base(l)+a] := s[t];
                  Dec(t);
                end;
          cal : begin
                  (*generate new block mark*)
                  s[t + 1] := base(l);
                  s[t +2] := b;
                  s[t+3] := p;
                  b := t + 1;
                  p := a
                end;
          int : t := t + a;
          jmp : p := a;
          jpc : begin
                  if s[t] = 0 then p := a;
                  Dec(t);
                end;
          wri: begin
                 WriteLn('wri: ' + IntToStr(s[t]));
                 Dec(t)
               end;
          opr : case a of (*operator*)
                  0: begin //return
                       t := b - 1;
                       p := s[t+3];
                       b := s[t+2]
                     end;
                  1: s[t] := -s[t]; //negation
                  2: begin
                       //Addition
                       Dec(t);
                       s[t] := s[t] + s[t+1]
                     end;
                  3: begin
                       //Subtraktion
                       Dec(t);
                       s[t] := s[t] - s[t+1]
                     end;
                  4: begin
                       //mul
                       Dec(t);
                       s[t] := s[t] * s[t+1]
                     end;
                  5: begin
                       //div
                       Dec(t);
                       s[t] := s[t] div s[t+1]
                     end;
                     //6 und 7 aus dem PL/0 interpreter von N.Wirth werden NICHT bentigt
                  8: begin
                       //Equal
                       Dec(t);
                       s[t] := ord(s[t] = s[t+1])
                     end;
                  9: begin
                       //unequal
                       Dec(t);
                       s[t] := ord(s[t] <> s[t+1])
                     end;
                  10:begin
                       //smaller
                       Dec(t);
                       s[t] := ord(s[t] < s[t+1])
                     end;
                  11:begin
                       //Bigger
                       Dec(t);
                       s[t] := ord(s[t] > s[t+1]);
                     end;
                  12:begin
                       //Bigerequal
                       Dec(t);
                       s[t] := ord(s[t] >= s[t+1]);
                     end;
                  13:begin
                       //Smallerequal
                       Dec(t);
                       s[t] := ord(s[t] <= s[t+1]);
                     end;
                else
                  WriteLn('Unknown Operand!');
                  p := 0;
                end;
        else
          WriteLn('Unknown opcode');
          p := 0;
        end
    until p = 0;
    WriteLn('Done...');
  end;

var
  F : File;
  FSize : Integer;

  procedure ShowInstructions;
  const
    OpCodeText : Array[TOpCode] of ShortString = ('lit', 'opr', 'lod', 'sto', 'cal', 'int', 'jmp', 'jpc', 'wri');
  var
    I : Integer;
  begin
    WriteLn('Instructions:');
    I := 0;
    while (I+1) * SizeOf(Instruction)<= FSize do
    begin
      {$R-}
      with Instructions^[I] do
        WriteLn(Format('%d: %s %d,%d', [i, OpCodeText[f], l, a]));
      {$R+}
      Inc(i)
    end
  end;

begin
  if not FileExists(ParamStr(1)) then Halt(0);

  AssignFile(F, ParamStr(1));
  FileMode := 0;
  Reset(F, 1);
  FSize := FileSize(F);
  GetMem(Instructions, FSize);
  BlockRead(F, Instructions^[0], FSize);
  if ParamStr(2) = '+i' then
    ShowInstructions;
  Emulate;
  FreeMem(Instructions, FSize);
  ReadLn;

  CloseFile(F)
end.
