unit errors;

interface
uses SysUtils, Symbols;

var
  Line : Integer;

resourcestring
  ERR_SCANNER_UNEXPECTED_CHAR = 'Error 0: Scanner: Unexpected Char Found in Stream';
  ERR_PARSE_EXPECTED = 'Error 1: Parser: %s expected, %s found instead';
  ERR_PARSE_UNALLOWED_STATEMENT = 'Error 2: Unallowed Statement';
  ERR_PARSE_WRONG_PROCEDURE_ENDED = 'Error 3: Procedureend %s expected, but %s found';
  ERR_PARSE_UNKNOWN_IDENT = 'Error 4: Unknown Identifier';
  ERR_PARSE_VAR_CONSTANT_EXPECTED = 'Error 5: Variable or Constant expected';
  ERR_PARSE_VAR_EXPECTED = 'Error 6: Variable expected';
  ERR_PARSE_PROCEDURE_EXPECTED = 'Error 7: Procedure expected';
  ERR_PARSE_NO_CONST_ALLOWED = 'Error 8: No Constant allowed here';

procedure Error(ErrorText : String);
procedure ErrorExpected(Expected : Array of TSymbol; Found : TSymbol);

implementation

procedure Error(ErrorText : String);
begin
  WriteLn(Format('%d: ' + ErrorText, [Line]));
  ReadLn;
  Halt(0)
end;

procedure ErrorExpected(Expected : Array of TSymbol; Found : TSymbol);
const
  eSymbols : Array[TSymbol] of ShortString = ('Unknown', 'Identifier', 'Integer', '+', '-', '*', '/', '=',
               '<', '>', '>=', '<=', '#',
               '(', ')', ',', '.', ';', ':=',
               'VAR', 'CONST', 'PROCEDURE', 'BEGIN', 'END', 'IF', 'THEN',
               'ELSEIF', 'ELSE', 'WHILE', 'DO', 'MODULE', 'WRITE', '!none!');

var
  ExpectedSymbols : String;
  I : Integer;
begin
  ExpectedSymbols := eSymbols[Expected[Low(Expected)]];
  for i := Low(Expected)+1 to High(Expected) do
    ExpectedSymbols := ExpectedSymbols + ', ' + eSymbols[Expected[i]];
  WriteLn(Format('%d: ' + ERR_PARSE_EXPECTED, [Line, ExpectedSymbols, eSymbols[Found]]));
  ReadLn;
  Halt(0)
end;

end.
