program parser;
{$APPTYPE CONSOLE}
uses
  sysutils,
  symbols in 'symbols.pas',
  scanner in 'scanner.pas',
  errors in 'errors.pas';

  procedure Expect(Expected : TSymbol);
  begin
    if Symbol <> Expected then
      ErrorExpected([Expected], Symbol)
  end;

  procedure Module;

    procedure StatementSequence;
    (*  StatementSeq  = [Statement {";" Statement}]*)
      procedure Statement;
      (*Statement     = ( Ident ":=" (Ident | Expression) |
                  Ident | (WRITE Expression) |
                  ( BEGIN StatementSeq END ) |
                  ( IF Condition THEN StatementSeq END
                    {ELSEIF Condition THEN StatementSeq END }
                    [ELSE StatementSeq END] ) |
                  ( WHILE Condition DO StatementSeq END ) | ";". *)
        procedure Expression;
        (*Expression      = [Vorzeichen] Term {Vorzeichen Term}.*)
          procedure Term;
          (*Term            = Factor {"*"|"/" Factor}.*)
            procedure Factor;
            (*Factor          = Ident | Zahl | ("(" Expression ")").*)
            begin (*Factor*)
              if (Symbol in [sIdent, sInteger]) then
                GetSym
              else if (Symbol = sOpenBracket) then
              begin
                GetSym;
                Expression;
                Expect(sCloseBracket);
                GetSym
              end
              else
                ErrorExpected([sIdent, sInteger, sOpenBracket], Symbol)
            end;  (*Factor*)

          begin(*Term*)
            Factor;
            while Symbol in [sStar, sSlash] do
            begin
              GetSym;
              Factor
            end
          end;(*Term*)
        begin (*Expresion*)
          if (Symbol in[sPlus, sMinus]) then
            GetSym;
          Term;
          while (Symbol in[sPlus, sMinus]) do
          begin
            GetSym;
            Term
          end
        end; (*Expression*)

        procedure Condition;
        (*Condition       = Expression ("="|"<"|">"|">="|"<="|"#") Expression.*)
        begin (*Condition*)
          Expression;
          if Symbol in [sEqual, sSmaller, sBigger, sBiggerEqual, sSmallerEqual, sUnequal] then
          begin
            GetSym;
            Expression
          end
          else
            ErrorExpected([sEqual, sSmaller, sBigger, sBiggerEqual, sSmallerEqual, sUnequal], Symbol)
        end; (*Condition*)

      begin (*Statement*)
        case Symbol of
          sIdent: begin
                    GetSym;
                    if Symbol = sBecomes then
                    begin
                      GetSym;
                      Expression;
                    end
                    else ; //Funktionsaufruf
                  end;
          sWrite: begin
                    GetSym;
                    Expression
                  end;
          sIf :   begin
                    GetSym;
                    Condition;
                    Expect(sThen);
                    GetSym;
                    StatementSequence;
                    Expect(sEnd);
                    GetSym;
                    while Symbol = sElseIf do
                    begin
                      GetSym;
                      Condition;
                      Expect(sThen);
                      GetSym;
                      StatementSequence;
                      Expect(sEnd);
                      GetSym
                    end;
                    if Symbol = sElse then
                    begin
                      GetSym;
                      StatementSequence;
                      Expect(sEnd);
                      GetSym
                    end
                  end;
          sWhile :
                  begin
                    GetSym;
                    Condition;
                    Expect(sDo);
                    GetSym;
                    StatementSequence;
                    Expect(sEnd);
                    GetSym
                  end;
          sBegin :
                  begin
                    GetSym;
                    StatementSequence;
                    Expect(sEnd);
                    GetSym
                  end;

        else
          (*Der Fehler darf getrost ignoriert werden*)
          //Error(ERR_PARSE_UNALLOWED_STATEMENT)
        end
      end; (*Statement*)

    begin (*Statement Sequence*)
      Statement;
      while Symbol = sSemiColon do
      begin
        GetSym;
        Statement
      end
    end; (*Statement Sequence*)

    procedure Declarations;
    (*
      VarSeq          = VAR VarDecl {";" VarDecl}";".
      ConstSeq        = CONST ConstDecl {";" ConstDecl}";".
      DeclSeq         = {VarSeq | ConstSeq | ProcedureDecl}.

    *)
      procedure ProcedureDecl;
      (*ProcedureDecl   = PROCEDURE Ident";"
                          DeclSeq BEGIN StatementSeq END Ident ";".*)
      begin
        Expect(sProcedure);
        GetSym;
        Expect(sIdent);
        GetSym;
        Expect(sSemiColon);
        GetSym;
        Declarations;
        Expect(sBegin);
        GetSym;
        StatementSequence;
        Expect(sEnd);
        GetSym;
        Expect(sIdent);
        GetSym;
        Expect(sSemiColon);
        GetSym
      end;

      procedure ConstDecl;
      (*ConstDecl       = Ident "=" Zahl.*)
      begin (*ConstDecl*)
        Expect(sIdent);
        GetSym;
        Expect(sEqual);
        GetSym;
        Expect(sInteger);
        GetSym
      end; (*ConstDecl*)

      procedure VarDecl;
      (*VarDecl         = Ident {"," Ident}.*)
      begin (*VarDecl*)
        Expect(sIdent);
        GetSym;
        while Symbol = sComma do
        begin
          GetSym;
          Expect(sIdent);
          GetSym
        end
      end; (*VarDecl*)

    begin (*Declarations*)
      while Symbol in [sVar, sConst, sProcedure] do
        case Symbol of
          sVar   : begin
                     GetSym;
                     VarDecl;
                     Expect(sSemiColon);
                     GetSym;
                     while Symbol = sIdent do
                     begin
                       VarDecl;
                       Expect(sSemiColon);
                       GetSym
                     end
                   end;
         sConst  : begin
                     GetSym;
                     ConstDecl;
                     Expect(sSemiColon);
                     GetSym;
                     while Symbol = sIdent do
                     begin
                       ConstDecl;
                       Expect(sSemiColon);
                       GetSym
                     end
                   end;
          sProcedure : ProcedureDecl;
        end
    end; (*Declarations*)
  (*Module        = MODULE Ident ";"
                    DeclSeq
                    BEGIN
                      StatementSeq
                    END Ident ".".*)
  begin (*Module*)
    Expect(sModule);
    GetSym;
    Expect(sIdent);
    GetSym;
    Expect(sSemiColon);
    GetSym;
    Declarations;
    Expect(sBegin);
    GetSym;
    StatementSequence;
    Expect(sEnd);
    GetSym;
    Expect(sIdent);
    GetSym;
    Expect(sSemiColon);
    GetSym;
    if Symbol <> sNone then
      WriteLn(Format('%d: Code after Module END is ignored!', [Line]))
  end (*Module*);

begin
  AssignFile(InFile, ParamStr(1));
  Reset(InFile, 1);
  //Ch mit leerzeichen initialisieren
  Ch := ' ';
  Line := 1;
  GetSym;
  Module;

  WriteLn('Done, no syntax errors detected...');
  ReadLn;
  CloseFile(InFile)
end.
