program scanner;
{$APPTYPE CONSOLE}
uses
  sysutils;

type
  TSymbol = (sUnknown, sIdent, sInteger, sPlus, sMinus, sStar, sSlash, sEqual,
             sSmaller, sBigger, sBiggerEqual, sSmallerEqual, sUnEqual,
             sOpenBracket, sCloseBracket, sComma, sDot, sSemiColon, sBecomes,
             sVar, sConst, sProcedure, sBegin, sEnd, sIf, sThen,
             sElseIf, sElse, sWhile, sDo, sModule, sNone);
const
  cSymbols : Array[TSymbol] of ShortString = ('', '', '', '+', '-', '*', '/', '=',
               '<', '>', '>=', '<=', '#',
               '(', ')', ',', '.', ';', ':=',
               'VAR', 'CONST', 'PROCEDURE', 'BEGIN', 'END', 'IF', 'THEN',
               'ELSEIF', 'ELSE', 'WHILE', 'DO', 'MODULE', '');

resourcestring
  ERR_SCANNER_UNEXPECTED_CHAR = 'Error 0: Scanner: Unexpected Char Found in Stream';


var
  InFile : File;
  Line : Integer;

  procedure Error(ErrorText : String);
  begin
    WriteLn(Format('%d: ' + ErrorText, [Line]));
    ReadLn;
    Halt(0)
  end;

var
  Ch : Char;
  Str : String;
  Symbol : TSymbol;

  procedure GetSym;
    procedure GetCh;
    begin
      if not EoF(InFile) then
        BlockRead(InFile, Ch, 1)
      else
        Ch := ' ';
      //Case in-sensitive
      Ch := UpCase(Ch);

      //Zeile erhhen?
      if Ch = #13 then Inc(Line);
      //Spezialzeichen filtern
      if Ord(Ch) < Ord(' ') then Ch := ' '
    end (*GetCh*);
  var
    I : TSymbol;
  begin
    //berspringen von Kommentaren
    while true do
    begin
      Str := '';
      Symbol := sNone;

      while (Ch = ' ') and not EoF(InFile) do
        GetCh;

      //Wir brauchen nichts machen, wenn das Dateiende erreicht ist
      if EoF(InFile) then
        Exit;

      case Ch of
        'A'..'Z','_': begin (*Ident/Reserved Word*)
                        while Ch in ['A'..'Z','_','0'..'9'] do
                        begin
                          Str := Str + Ch;
                          GetCh
                        end;
                        Symbol := sIdent;

                        for i := sUnknown to sNone do
                          if Str = cSymbols[i] then
                          begin
                            Symbol := i;
                            Break
                          end;
                        Exit
                      end; (*Ident/Reserved Word*)
          ';','+','-','=','#',',','.':
                      begin (*Symbole die nur aus 1 Zeichen bestehen knnen*)
                        Str := Ch;
                        Symbol := sUnknown;
                        for i := sUnknown to sNone do
                          if Str = cSymbols[i] then
                          begin
                            Symbol := i;
                            Break
                          end;
                        GetCh;
                        Exit
                      end; (*Symbole die nur aus 1 Zeichen bestehen knnen*)
          ':','<','>': (* Zeichen die ein nachfolgendes Zeichen haben knnen(in diesm Falle ein = )*)
                      begin
                        Str := Ch;
                        GetCh;
                        if Ch = '=' then Str := Str + Ch;
                        GetCh;
                        Symbol := sUnknown;
                        for i := sUnknown to sNone do
                          if Str = cSymbols[i] then
                          begin
                            Symbol := i;
                            Break
                          end;
                        Exit
                      end;
          '(', ')':   begin (*Klammern oder Kommentar*)
                        Str := Ch;
                        GetCh;
                        if (Str='(') and (Ch = '*') then
                        begin
                          //Kommentar entdeckt
                          GetCh;
                          while True do
                          begin
                            GetCh;
                            if Ch = '*' then
                            begin
                              GetCh;
                              if Ch = ')' then
                              begin
                                GetCh;
                                Break
                              end
                            end
                            else if EoF(InFile) then
                              Break
                          end
                        end else
                        if Str = '(' then
                        begin
                          Symbol := sOpenBracket;
                          Exit
                        end
                        else
                        begin
                          Symbol := sCloseBracket;
                          Exit
                        end
                      end;
          '0'..'9','$': (*Zahlen*)
                      begin
                        Symbol := sInteger;
                        Str := Ch;
                        GetCh;
                        if (Str = '$') then
                        begin
                          //HexZahl
                          while Ch in ['0'..'9','A'..'F'] do
                          begin
                            Str := Str + Ch;
                            GetCh;
                          end;
                          Exit
                        end
                        else
                        begin
                          //NormaleZahl
                          while Ch in ['0'..'9'] do
                          begin
                            Str := Str + Ch;
                            GetCh;
                          end;
                          Exit
                        end
                      end; (*Zahlen*)

      else
        Error(ERR_SCANNER_UNEXPECTED_CHAR)
      end;
      Assert(Symbol <> SUnknown);
    end
  end (*GetSym*);

begin
  AssignFile(InFile, ParamStr(1));
  Reset(InFile, 1);
  //Ch mit leerzeichen initialisieren
  Ch := ' ';
  Line := 1;
  GetSym;
  while Symbol <> sNone do
  begin
    WriteLn(Str);
    GetSym
  end;
  WriteLn('Done...');
  ReadLn;
  CloseFile(InFile)
end.
